package ratRace.model;

import java.util.Observable;


/** The Model is a facade object representing all components
 *  of the model to other components of the system.  These components
 *  are a maze and a rat that moves around within the maze.
 *  The Model is also responsible for keeping track of the amount of
 *  time that has passed during the game and whether the game is paused.
 *  The Model receives events (from the Controller) telling it where to send
 *  the rat, that time has passed, and when to pause or unpause the game.
 *  When its state changes, the model notifies its obverver(s) (typically the View).
 * 
 * @author Theodore Norvell
 *
 */
public class Model extends Observable {
	private Maze maze ;
	private Rat rat ;
	private boolean paused = true ;
	private long timeOfLastPulse = System.currentTimeMillis() ;
	private boolean gameOver;
	private long elapsedTime ;
	
	public Model(int rows, int cols) {
		this.maze =  new Maze(rows, cols) ;
		this.rat = new Rat( maze ) ;
		this.maze.buildInitialWalls( 40 ) ;
	}
	
	public boolean gameOver( ) { return gameOver ; }
	
	public int getMazeWidth()  { return maze.getMazeWidth() ; }
	
	public int getMazeHeight() { return maze.getMazeHeight() ; }
	
	public void pointRatAt(double i, double j) {
		rat.pointAt(i,j) ;
	}
	
	public void draw( DrawingAdapter drawingAdapter ) {
		String message = ""+ (elapsedTime/1000.0) ;
		drawingAdapter.putMessage( message, gameOver ) ;
		
		maze.draw( drawingAdapter ) ;
		rat.draw(drawingAdapter ) ;
	}
	
	public void setPaused( boolean paused ) {
		this.paused = paused ;
		timeOfLastPulse = System.currentTimeMillis() ;
		setChanged() ;
		notifyObservers() ;
	}
	
	public boolean getPaused() { return paused ; }
	
	public void pulse() {
		//System.out.println("Model.pulse" ) ;
		long time = System.currentTimeMillis() ;
		long delta_t = timeOfLastPulse = time - timeOfLastPulse ;
		timeOfLastPulse = time ;
		if( paused ) return ;
		if( ! gameOver ) this.elapsedTime += delta_t ;
		maze.pulse( delta_t ) ;
		rat.pulse( delta_t ) ;
		if( rat.getBlock_i(rat.get_i())==maze.goal_i()
				&& rat.getBlock_j(rat.get_j())==maze.goal_j() ) {
			gameOver = true ;
		}
		//System.out.println("Model.pulse notifies observers" ) ;
		setChanged() ;
		notifyObservers() ;
	}
}
