package ratRace.model;


class Rat {
	private final Maze maze ;
	
	// The location in the maze as a real number
	// The rat starts in the middle of the upper left block.
	private double i = 0.5, j = 0.5 ;
	
	// How big is the rat as a circle.
	private final double radius = 0.25 ;
	
	// Where is the mouse.
	private double mouse_i = 100, mouse_j = 100 ;

	private void checkInvariant() {
		if( i < radius  ) throw new AssertionError() ;
		if( i > maze.getMazeHeight()-radius ) throw new AssertionError() ;
		if( j < radius  ) throw new AssertionError() ;
		if( j > maze.getMazeWidth()-radius ) throw new AssertionError() ;
	}
	
	Rat( Maze maze ) {
		this.maze = maze ;
		this.maze.setRat(this) ;
	}
	
	/** Where is the rat? */
	double get_i() { return i ; }
	
	/** Where is the rat? */
	double get_j() { return j ; }
	
	/** What block is the rat in? */ 
	int getBlock_i( double i) { return (int) Math.floor(i) ; }
	
	/** What block is the rat in? */ 
	int getBlock_j(double j) { return (int) Math.floor(j) ; }
	
	/** React to a change in time */
	void pulse(long delta_t) {

		double dir_i = mouse_i - i;
		double dir_j = mouse_j - j;
		double hypot = Math.sqrt(dir_i*dir_i+dir_j*dir_j) ;
		// For each block, the mouse is away from the
		// rat, the velocity is increased by 0.001 blocks per
		// millisecond (i.e. 1 block per second).
		double velocity = hypot * 0.001 ;
		// Cap the velocity at 10 blocks per second.
		velocity = Math.min(velocity, 0.01) ;
		
		// Compute the new i and j values
		double distToMove = Math.min(delta_t * velocity, 1.0) ;
		double scale = distToMove/Math.max(hypot,0.0001) ;
		double delta_i = scale * dir_i ;
		double delta_j = scale * dir_j ;
		double new_i = i + delta_i ;
		double new_j = j + delta_j ;
		
		if( delta_i > 1 || delta_i < -1 ) throw new AssertionError() ;
		if( delta_j > 1 || delta_j < -1 ) throw new AssertionError() ;
		
		int block_i = getBlock_i(i) ; 	
		int block_j = getBlock_j(j) ;
		
		// Since the rat is travelling a distance of less than 1
		// block, we only need to check one vertical and one
		// horizontal wall.
		
		if( delta_i < 0 ) { // Going up
			// The rat is considered blocked if the new location would put
			// any part of it above block_i and there is a wall
			// above the rat.
			// (Strictly speaking this is a conservative check, 
			// because the trajectory of the rat could miss the wall
			// in which case it should not be blocked.)
			boolean blocked = new_i-radius < block_i 
				&& maze.hasHWall(block_i, block_j);
			if( blocked ) { i = block_i + radius ; }
			else { i = Math.max(new_i, radius ) ; } }
		
		else { // Going down or horizontal
			boolean blocked = new_i+radius > block_i+1
					&& maze.hasHWall(block_i+1, block_j);
			if( blocked ) { i = block_i+1 - radius ; }
			else { i = Math.min(new_i, maze.getMazeHeight()-radius) ; } }
		
		if( delta_j<0 ) { // Going left.
			boolean blocked = new_j-radius < block_j 
					&& maze.hasVWall(block_i,  block_j) ;
			if( blocked ) { j = block_j + radius ; }
			else { j = Math.max(new_j, radius) ; } }
		
		else { // Going right or vertical
			boolean blocked = new_j+radius > block_j+1
					&& maze.hasVWall(block_i, block_j+1 ) ;
			if( blocked ) { j = block_j+1 - radius ; }
			else { j = Math.min(new_j, maze.getMazeWidth()-radius) ; } }

		checkInvariant() ;
	}

	void pointAt(double mouse_i, double mouse_j) {
		this.mouse_i = mouse_i ;
		this.mouse_j = mouse_j ;
	}

	public void draw(DrawingAdapter drawingAdapter) {
		drawingAdapter.drawRat( i, j, radius ) ;
		
	}
	
}
