package ratRace.viewSwing;

import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionAdapter;
import java.awt.geom.Point2D;
import java.util.Set;
import java.util.HashSet;
import java.util.Observable;
import java.util.Observer;
import javax.swing.JPanel;

import ratRace.model.DrawingAdapter;
import ratRace.model.Model;
import ratRace.mvcInterfaces.ViewI;
import ratRace.mvcInterfaces.ViewListenerI;


/** The SwingView is responsible for drawing a visual representation of the model;
 * it also translates mouse moved events to world coordinates and sends them to
 * its observer(s).
 * 
 * @author Theodore Norvell
 *
 */
public class SwingView extends JPanel implements Observer, ViewI {
	private static final long serialVersionUID = -1066006314191861658L;
	private final Model model ;
	private final int modelWidth ;
	private final int modelHeight ;
	private final Set<ViewListenerI> myListeners 
	    = new HashSet<ViewListenerI>() ;
	
	public SwingView( Model model ) {
		this.model = model ;
		modelWidth = model.getMazeWidth();
		modelHeight = model.getMazeHeight();
		model.addObserver( this ) ;
		addMouseMotionListener(new MyMouseListener()) ;
		timer.setRepeats(true) ;
		timer.start() ;
		repaint() ;
	}
	
	public void addListener( ViewListenerI listener) {
		myListeners.add( listener ) ;
	}
	
	@Override
	public void update(Observable arg0, Object arg1) {
		repaint() ;
	}

	public void paintComponent( Graphics g ) {
		super.paintComponent(g) ;
		if( !(g instanceof Graphics2D) )
			throw new AssertionError("Needs Graphics2D support." ) ;
		Graphics2D g2d = (Graphics2D) g.create() ;
		int width = getWidth();
		int height = getHeight() ;
		
		Transform xform
			= new Transform( width, height,
					         model.getMazeWidth(), model.getMazeWidth() ) ;
		
		DrawingAdapter da = new SwingDrawingAdapter( g2d, xform, width, height ) ;
		
		model.draw( da ) ; 
	}	
	public void terminate() {
		timer.stop() ;
	}
	
	// It is important to use a Swing timer here
	private javax.swing.Timer timer = new javax.swing.Timer(50,
			new ActionListener() {
				@Override
				public void actionPerformed(ActionEvent arg0) {
					for( ViewListenerI listener : myListeners ) {
						listener.pulse() ;
					} }});
	
	class MyMouseListener extends MouseMotionAdapter {
		@Override
		public void mouseMoved(MouseEvent e) {
			double x = e.getX() ;
			double y = e.getY();
			Transform xform
			= new Transform( getWidth(), getHeight(),
			         model.getMazeWidth(), model.getMazeWidth() ) ;
			Point2D p = xform.viewToWorld(x, y) ;
			for( ViewListenerI listener : myListeners ) {
				listener.mouseMovedTo(p.getX(), p.getY()) ;
			}
		}
	}
}
